#!/usr/bin/python3

"""Visitor for substitution"""

from __future__ import annotations
from abc import abstractmethod
from collections.abc import Set
from itertools import count, filterfalse
from typing import TypeVar
from .. import terms
from ..errors import CollisionError
from . import Visitor

__all__ = (
    "SubstitutingVisitor",
    "RenamingSubstitutingVisitor",
    "CountingSubstitutingVisitor"
)

V = TypeVar("V")


class SubstitutingVisitor(Visitor["terms.Term[V]", V]):
    """
    Visitor which replaces a free Variable with another term
    Raises a CollisionError if a free variable gets bound
    """

    variable: V

    value: terms.Term[V]

    free_variables: Set[V]

    bound_variables: dict[V, int]

    __slots__ = (
        "variable",
        "value",
        "free_variables",
        "bound_variables"
    )

    def __init__(self, variable: V, value: terms.Term[V]) -> None:
        self.variable = variable
        self.value = value
        self.free_variables = value.free_variables()
        # store number of binds since a variable can be bound multiple times
        self.bound_variables = {}

    def bind_variable(self, name: V) -> None:
        """mark a variable as bound"""
        self.bound_variables[name] = self.bound_variables.get(name, 0) + 1

    def unbind_variable(self, name: V) -> None:
        """mark a variable as not bound"""
        number = self.bound_variables[name]
        if number == 1:
            del self.bound_variables[name]
        else:
            self.bound_variables[name] = number - 1

    def visit_variable(self, variable: terms.Variable[V]) -> terms.Term[V]:
        """visit a Variable term"""
        if variable.name != self.variable:
            return variable
        collisions = self.free_variables & self.bound_variables.keys()
        if collisions:
            raise CollisionError("free variables in value are bound in term", collisions)
        return self.value

    def visit_abstraction(self, abstraction: terms.Abstraction[V]) -> terms.Abstraction[V]:
        """visit an Abstraction term"""
        if abstraction.bound == self.variable:
            return abstraction
        self.bind_variable(abstraction.bound)
        try:
            return terms.Abstraction(
                abstraction.bound,
                abstraction.body.accept(self)
            )
        finally:
            # allow reuse of this visitor, even on error
            self.unbind_variable(abstraction.bound)

    def visit_application(self, application: terms.Application[V]) -> terms.Application[V]:
        """visit an Application term"""
        return terms.Application(
            application.abstraction.accept(self),
            application.argument.accept(self)
        )


class RenamingSubstitutingVisitor(Visitor["terms.Term[V]", V]):
    """
    Visitor which replaces a free Variable with another term
    Renames bound variables if a free variable gets bound
    """

    variable: V

    value: terms.Term[V]

    __slots__ = (
        "variable",
        "value"
    )

    @abstractmethod
    def prevent_collision(self, abstraction: terms.Abstraction[V]) -> terms.Abstraction[V]:
        """prevent collisions by renaming bound variables"""
        raise NotImplementedError()

    def visit_variable(self, variable: terms.Variable[V]) -> terms.Term[V]:
        """visit a Variable term"""
        if variable.name != self.variable:
            return variable
        return self.value

    def visit_abstraction(self, abstraction: terms.Abstraction[V]) -> terms.Abstraction[V]:
        """visit an Abstraction term"""
        if abstraction.bound == self.variable:
            return abstraction
        renamed = self.prevent_collision(abstraction)
        return terms.Abstraction(
            renamed.bound,
            renamed.body.accept(self)
        )

    def visit_application(self, application: terms.Application[V]) -> terms.Application[V]:
        """visit an Application term"""
        return terms.Application(
            application.abstraction.accept(self),
            application.argument.accept(self)
        )


class CountingSubstitutingVisitor(RenamingSubstitutingVisitor[str]):
    """
    Visitor which replaces a free Variable with another term
    Renames bound variables if a free variable gets bound by appending a number
    """

    free_variables: Set[str]

    __slots__ = ("free_variables",)

    def __init__(self, variable: str, value: terms.Term[str]) -> None:
        self.variable = variable
        self.value = value
        self.free_variables = value.free_variables()

    def prevent_collision(self, abstraction: terms.Abstraction[str]) -> terms.Abstraction[str]:
        """prevent collisions by renaming bound variables"""
        if abstraction.bound in self.free_variables:
            used_variables = abstraction.body.bound_variables() \
                | abstraction.free_variables() \
                | self.free_variables
            candidates = map(lambda i: f"{abstraction.bound}{i}", count(1))
            variable = next(filterfalse(lambda v: v in used_variables, candidates))
            return terms.Abstraction(
                variable,
                abstraction.body.substitute(abstraction.bound, terms.Variable(variable))
            )
        else:
            return abstraction



#!/usr/bin/python3

"""Visitor for normalisation"""

from __future__ import annotations
from collections.abc import Iterator
from typing import TypeVar
from .. import terms
from . import Visitor
from .substitution import CountingSubstitutingVisitor

__all__ = (
    "BetaNormalisingVisitor",
)

V = TypeVar("V")


class BetaNormalisingVisitor(Visitor[Iterator[terms.Term[str]], str]):
    """
    Visitor which transforms a term into its beta normal form,
    yielding intermediate results until it is reached
    """

    __slots__ = ()

    def skip_intermediate(self, term: terms.Term[str]) -> terms.Term[str]:
        """return the beta normal form directly"""
        result = term
        for intermediate in term.accept(self):
            result = intermediate
        return result

    def visit_variable(self, variable: terms.Variable[str]) -> Iterator[terms.Variable[str]]:
        """visit a Variable term"""
        return iter(())

    def visit_abstraction(self, abstraction: terms.Abstraction[str]) -> Iterator[terms.Abstraction[str]]:
        """visit an Abstraction term"""
        results = abstraction.body.accept(self)
        return map(lambda b: terms.Abstraction(abstraction.bound, b), results)

    def visit_application(self, application: terms.Application[str]) -> Iterator[terms.Term[str]]:
        """visit an Application term"""
        match application.abstraction:
            # normal order dictates we reduce leftmost outermost redex first
            case terms.Abstraction(bound, body):
                reduced = body.accept(CountingSubstitutingVisitor(bound, application.argument))
                yield reduced
                yield from reduced.accept(self)
            case _:
                # try to reduce the abstraction until this is a redex
                abstraction = application.abstraction
                for transformation in application.abstraction.accept(self):
                    yield terms.Application(transformation, application.argument)
                    match transformation:
                        case terms.Abstraction(bound, body):
                            reduced = body.accept(
                                CountingSubstitutingVisitor(bound, application.argument)
                            )
                            yield reduced
                            yield from reduced.accept(self)
                            return
                        case _:
                            abstraction = transformation
                # no redex, continue with argument
                transformations = application.argument.accept(self)
                yield from map(lambda a: terms.Application(abstraction, a), transformations)



#!/usr/bin/python3

"""Visitors for performing operations on Terms"""

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TypeVar, Generic
from .. import terms

__all__ = (
    "Visitor",
    "BottomUpVisitor",
    "substitution",
    "normalisation",
    "walking"
)

T = TypeVar("T")
V = TypeVar("V")


class Visitor(ABC, Generic[T, V]):
    """
    ABC for Visitors visiting Terms
    The visitor is responsible to visit child terms
    """

    __slots__ = ()

    def visit(self, term: terms.Term[V]) -> T:
        """visit a term"""
        return term.accept(self)

    @abstractmethod
    def visit_variable(self, variable: terms.Variable[V]) -> T:
        """visit a Variable term"""
        raise NotADirectoryError()

    @abstractmethod
    def visit_abstraction(self, abstraction: terms.Abstraction[V]) -> T:
        """visit an Abstraction term"""
        raise NotImplementedError()

    @abstractmethod
    def visit_application(self, application: terms.Application[V]) -> T:
        """visit an Application term"""
        raise NotImplementedError()


class BottomUpVisitor(Visitor[T, V]):
    """ABC for visitors which visit child terms first"""

    __slots__ = ()

    def visit_abstraction(self, abstraction: terms.Abstraction[V]) -> T:
        """visit an Abstraction term"""
        return self.ascend_abstraction(
            abstraction,
            abstraction.body.accept(self)
        )

    def visit_application(self, application: terms.Application[V]) -> T:
        """visit an Application term"""
        return self.ascend_application(
            application,
            application.abstraction.accept(self),
            application.argument.accept(self)
        )

    @abstractmethod
    def ascend_abstraction(self, abstraction: terms.Abstraction[V], body: T) -> T:
        """visit an Abstraction term after visiting its body"""
        raise NotImplementedError()

    @abstractmethod
    def ascend_application(self, application: terms.Application[V], abstraction: T, argument: T) -> T:
        """visit an Application term after visiting its abstraction and argument"""
        raise NotImplementedError()



#!/usr/bin/python3

"""Implementation of pairs"""

from . import Variable, Abstraction, Application
from .logic import TRUE, FALSE

__all__ = (
    "PAIR",
    "FIRST",
    "SECOND",
    "NIL",
    "NULL"
)

PAIR = Abstraction.curried(
    ("x", "y", "f"),
    Application.with_arguments(
        Variable("f"),
        (Variable("x"), Variable("y"))
    )
)

FIRST = Abstraction("p", Application(Variable("p"), TRUE))

SECOND = Abstraction("p", Application(Variable("p"), FALSE))

NIL = Abstraction("x", TRUE)

NULL = Abstraction(
    "p",
    Application(
        Variable("p"),
        Abstraction.curried(
            ("x", "y"),
            FALSE
        )
    )
)



#!/usr/bin/python3

"""Implementations of boolean values and logical operators"""

from . import Variable, Abstraction, Application

__all__ = (
    "TRUE",
    "FALSE",
    "AND",
    "OR",
    "NOT",
    "IF_THEN_ELSE"
)

TRUE = Abstraction.curried(("x", "y"), Variable("x"))
FALSE = Abstraction.curried(("x", "y"), Variable("y"))

AND = Abstraction.curried(
    ("p", "q"),
    Application.with_arguments(Variable("p"), (Variable("q"), Variable("p")))
)
OR = Abstraction.curried(
    ("p", "q"),
    Application.with_arguments(Variable("p"), (Variable("p"), Variable("q")))
)
NOT = Abstraction("p", Application.with_arguments(Variable("p"), (FALSE, TRUE)))

IF_THEN_ELSE = Abstraction.curried(
    ("p", "a", "b"),
    Application.with_arguments(Variable("p"), (Variable("a"), Variable("b")))
)



#!/usr/bin/python3

"""Lambda Terms"""

from __future__ import annotations
from abc import abstractmethod
from collections.abc import Sequence, Set, Iterable, Iterator
from dataclasses import dataclass
from typing import TypeVar
from .. import visitors
from ..errors import CollisionError
from ..visitors import substitution, walking

__all__ = (
    "Term",
    "Variable",
    "Abstraction",
    "Application",
    "arithmetic",
    "logic",
    "pairs"
)

T = TypeVar("T")
V = TypeVar("V")


class Term(Iterable["Term[V]"]):
    """ABC for Lambda terms"""

    __slots__ = ("__weakref__",)

    def __iter__(self) -> Iterator[Term[V]]:
        return self.accept(walking.DepthFirstVisitor())

    @abstractmethod
    def free_variables(self) -> Set[V]:
        """return the free variables of this Term"""
        raise NotImplementedError()

    @abstractmethod
    def bound_variables(self) -> Set[V]:
        """return the bound variables of this Term"""
        raise NotImplementedError()

    @abstractmethod
    def is_beta_normal_form(self) -> bool:
        """return if this Term is in beta-normal form"""
        raise NotImplementedError()

    @abstractmethod
    def accept(self, visitor: visitors.Visitor[T, V]) -> T:
        """accept a visitor by calling his corresponding method"""
        raise NotImplementedError()

    def substitute(self, variable: V, value: Term[V]) -> Term[V]:
        """substitute a free variable with a Term, possibly raising a CollisionError"""
        return self.accept(substitution.SubstitutingVisitor(variable, value))

    def is_combinator(self) -> bool:
        """return if this Term has no free variables"""
        return not self.free_variables()


@dataclass(unsafe_hash=True, slots=True)
class Variable(Term[V]):
    """Term consiting of a Variable"""

    name: V

    @classmethod
    def with_valid_name(cls, name: V) -> Variable[V]:
        """create an instance with a valid name or raise ValueError"""
        string = str(name)
        if not string:
            raise ValueError("empty string representation")
        for character in string:
            if character in "().λ" or character.isspace():
                raise ValueError(f"invalid character: '{character}'")
        return cls(name)

    def __str__(self) -> str:
        return str(self.name)

    def free_variables(self) -> Set[V]:
        """return the free variables of this Term"""
        return {self.name}

    def bound_variables(self) -> Set[V]:
        """return the bound variables of this Term"""
        return set()

    def is_beta_normal_form(self) -> bool:
        """return if this Term is in beta-normal form"""
        return True

    def accept(self, visitor: visitors.Visitor[T, V]) -> T:
        """accept a visitor by calling his corresponding method"""
        return visitor.visit_variable(self)


@dataclass(unsafe_hash=True, slots=True)
class Abstraction(Term[V]):
    """Term consisting of a lambda abstraction"""

    bound: V

    body: Term[V]

    @classmethod
    def curried(cls, variables: Sequence[V], body: Term[V]) -> Abstraction[V]:
        """create an Abstraction binding multiple variables"""
        match variables:
            case [*variables, inner]:
                term = cls(inner, body)
                for variable in reversed(variables):
                    term = cls(variable, term)
                return term
            case _:
                raise ValueError("no variables to bind")

    def __str__(self) -> str:
        return f"(λ{self.bound}.{self.body})"

    def free_variables(self) -> Set[V]:
        """return the free variables of this Term"""
        return self.body.free_variables() - {self.bound}

    def bound_variables(self) -> Set[V]:
        """return the bound variables of this Term"""
        return self.body.bound_variables() | {self.bound}

    def is_beta_normal_form(self) -> bool:
        """return if this Term is in beta-normal form"""
        return self.body.is_beta_normal_form()

    def alpha_conversion(self, new: V) -> Abstraction[V]:
        """rename the bound variable"""
        if new == self.bound:
            return self
        elif new not in self.body.free_variables():
            return Abstraction(
                new,
                self.body.substitute(self.bound, Variable(new))
            )
        raise CollisionError("new variable would bind free variable in body", (new,))

    def eta_reduction(self) -> Term[V]:   # type: ignore[return]
        """remove a useless abstraction"""
        match self.body:
            case Application(f, Variable(x)) if x == self.bound and x not in f.free_variables():
                return f
            case _:
                # mypy detects missing returns
                # because of https://github.com/python/mypy/issues/12534
                raise ValueError("abstraction is not useless")

    def accept(self, visitor: visitors.Visitor[T, V]) -> T:
        """accept a visitor by calling his corresponding method"""
        return visitor.visit_abstraction(self)

    def replace(self, *, bound: V | None = None, body: Term[V] | None = None) -> Abstraction[V]:
        """return a copy with specific attributes replaced"""
        return Abstraction(
            self.bound if bound is None else bound,
            self.body if body is None else body
        )


@dataclass(unsafe_hash=True, slots=True)
class Application(Term[V]):
    """Term consisting of an application"""

    abstraction: Term[V]

    argument: Term[V]

    @classmethod
    def with_arguments(cls, abstraction: Term[V], arguments: Sequence[Term[V]]) -> Application[V]:
        """create an Application applying the abstraction to multiple arguments"""
        match arguments:
            case [first, *rest]:
                term = cls(abstraction, first)
                for argument in rest:
                    term = cls(term, argument)
                return term
            case _:
                raise ValueError("no arguments to apply abstraction to")

    def __str__(self) -> str:
        return f"({self.abstraction} {self.argument})"

    def free_variables(self) -> Set[V]:
        """return the free variables of this Term"""
        return self.abstraction.free_variables() | self.argument.free_variables()

    def bound_variables(self) -> Set[V]:
        """return the bound variables of this Term"""
        return self.abstraction.bound_variables() | self.argument.bound_variables()

    def is_redex(self) -> bool:
        """return if this term can be reduced"""
        return isinstance(self.abstraction, Abstraction)

    def is_beta_normal_form(self) -> bool:
        """return if this Term is in beta-normal form"""
        return not self.is_redex() \
            and self.abstraction.is_beta_normal_form() \
            and self.argument.is_beta_normal_form()

    def beta_reduction(self) -> Term[V]:
        """remove an abstraction"""
        match self.abstraction:
            case Abstraction(bound, body):
                return body.substitute(bound, self.argument)
            case _:
                raise ValueError("can not perform reduction without known Abstraction")

    def accept(self, visitor: visitors.Visitor[T, V]) -> T:
        """accept a visitor by calling his corresponding method"""
        return visitor.visit_application(self)

    def replace(self, *, abstraction: Term[V] | None = None, argument: Term[V] | None = None) -> Application[V]:
        """return a copy with specific attributes replaced"""
        return Application(
            self.abstraction if abstraction is None else abstraction,
            self.argument if argument is None else argument
        )



#!/usr/bin/python3

"""Errors raised by Term operations"""

from __future__ import annotations
from collections.abc import Collection
from typing import TypeVar, Generic

__all__ = (
    "CollisionError",
)

V = TypeVar("V")


class CollisionError(ValueError, Generic[V]):
    """Exception thrown when a variable already exists"""

    message: str

    collisions: Collection[V]

    __match_args__ = __slots__ = ("message", "collisions")

    def __init__(self, message: str, collisions: Collection[V]) -> None:
        super().__init__(message, collisions)
        self.message = message
        self.collisions = collisions

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.message!r}, {self.collisions!r})"

    def __str__(self) -> str:
        collisions = ", ".join(map(str, self.collisions))
        return f"[collisions: {collisions}] {self.message}"



#!/usr/bin/python3

"""Visitor for walking terms"""

from __future__ import annotations
from collections.abc import Iterator
from typing import TypeVar
from .. import terms
from . import BottomUpVisitor

__all__ = (
    "DepthFirstVisitor",
)

V = TypeVar("V")


class DepthFirstVisitor(BottomUpVisitor[Iterator["terms.Term[V]"], V]):
    """Visitor yielding subterms depth first"""

    def visit_variable(self, variable: terms.Variable[V]) -> Iterator[terms.Term[V]]:
        """visit a Variable term"""
        yield variable

    def ascend_abstraction(
        self,
        abstraction: terms.Abstraction[V],
        body: Iterator[terms.Term[V]]
    ) -> Iterator[terms.Term[V]]:
        """visit an Abstraction term after visiting its body"""
        yield from body
        yield abstraction

    def ascend_application(
        self,
        application: terms.Application[V],
        abstraction: Iterator[terms.Term[V]],
        argument: Iterator[terms.Term[V]]
    ) -> Iterator[terms.Term[V]]:
        """visit an Application term after visiting its abstraction and argument"""
        yield from abstraction
        yield from argument
        yield application



#!/usr/bin/python3

"""Implementations of natural numbers and arithmetic operators"""

from . import Term, Variable, Abstraction, Application

__all__ = (
    "SUCCESSOR",
    "PREDECESSOR",
    "ADD",
    "SUBTRACT",
    "MULTIPLY",
    "POWER",
    "number"
)

SUCCESSOR = Abstraction.curried(
    ("n", "f", "x"),
    Application(
        Variable("f"),
        Application.with_arguments(
            Variable("n"),
            (Variable("f"), Variable("x"))
        )
    )
)

PREDECESSOR = Abstraction.curried(
    ("n", "f", "x"),
    Application.with_arguments(
        Variable("n"),
        (
            Abstraction.curried(
                ("g", "h"),
                Application(
                    Variable("h"),
                    Application(
                        Variable("g"),
                        Variable("f")
                    )
                )
            ),
            Abstraction("u", Variable("x")),
            Abstraction("u", Variable("u")),
        )
    )
)

ADD = Abstraction.curried(
    ("m", "n", "f", "x"),
    Application.with_arguments(
        Variable("m"),
        (
            Variable("f"),
            Application.with_arguments(
                Variable("n"),
                (Variable("f"), Variable("x"))
            )
        )
    )
)

SUBTRACT = Abstraction.curried(
    ("m", "n"),
    Application.with_arguments(
        Variable("n"),
        (PREDECESSOR, Variable("m"))
    )
)

MULTIPLY = Abstraction.curried(
    ("m", "n", "f"),
    Application(
        Variable("m"),
        Application(
            Variable("n"),
            Variable("f")
        )
    )
)

POWER = Abstraction.curried(
    ("b", "e"),
    Application(
        Variable("e"),
        Variable("b")
    )
)


def number(n: int) -> Abstraction[str]:
    """return n encoded in lambda terms"""
    if n < 0:
        raise ValueError("number is not natural")
    f = Variable("f")
    body: Term[str] = Variable("x")
    for _ in range(n):
        body = Application(f, body)
    return Abstraction.curried(
        ("f", "x"),
        body
    )



#!/usr/bin/python3

"""Implementation of the Lambda calculus"""

from .terms import Variable, Abstraction, Application

__version__ = "1.10.1"
__author__  = "Eric Niklas Wolf"
__email__   = "eric_niklas.wolf@mailbox.tu-dresden.de"
__all__ = (
    "Variable",
    "Abstraction",
    "Application",
    "terms",
    "visitors",
    "errors"
)
